import tkinter as tk
from tkinter import ttk, filedialog, messagebox, simpledialog, Toplevel
import os
import json
import subprocess
import re
import struct
from typing import List, Dict, Tuple, Optional, Any

class WADParser:
    def __init__(self, wad_path: str):
        self.wad_path = wad_path
        self.levels = {}
        self.wad_data = None
        self.lumps = []
        self.parse_wad()
        
    def parse_wad(self):
        """Parse the WAD file to extract level markers and associated information."""
        try:
            with open(self.wad_path, 'rb') as f:
                self.wad_data = f.read()
                
                # WAD header: 4 bytes for type, 4 bytes for number of entries, 4 bytes for directory offset
                header = self.wad_data[:12]
                wad_type = header[:4].decode('ascii')
                num_lumps = struct.unpack("<I", header[4:8])[0]
                dir_offset = struct.unpack("<I", header[8:12])[0]
                
                # Read the directory
                self.lumps = []
                for i in range(num_lumps):
                    entry_offset = dir_offset + i * 16
                    lump_entry = self.wad_data[entry_offset:entry_offset + 16]
                    offset, size, name_bytes = struct.unpack("<ii8s", lump_entry)
                    name = name_bytes.decode('ascii', errors='replace').rstrip('\0')
                    self.lumps.append({
                        "offset": offset,
                        "size": size,
                        "name": name
                    })
                
                # Group lumps by level marker
                current_level = None
                for i, lump in enumerate(self.lumps):
                    if self.is_level_marker(lump["name"]):
                        current_level = lump["name"]
                        self.levels[current_level] = {"lumps": [], "index": i}
                    elif current_level is not None:
                        self.levels[current_level]["lumps"].append(lump)
                        
                        # If we find another level marker, stop adding to the current level
                        if i + 1 < len(self.lumps) and self.is_level_marker(self.lumps[i + 1]["name"]):
                            current_level = None
                
                # Extract additional info (for example, the THINGS lump)
                for level, data in self.levels.items():
                    things_lump = next((l for l in data["lumps"] if l["name"] == "THINGS"), None)
                    if things_lump:
                        # In classic Doom, each thing record is 10 bytes
                        things_count = things_lump["size"] // 10
                        data["thing_count"] = things_count
                
        except Exception as e:
            print(f"Error parsing WAD file: {e}")
    
    def get_lump_data(self, lump):
        """Extract binary data for a specific lump."""
        if not self.wad_data:
            return b''
        return self.wad_data[lump["offset"]:lump["offset"] + lump["size"]]
    
    def get_level_lumps(self, level_name):
        """Get the important lumps for a level: VERTEXES, LINEDEFS, THINGS."""
        if level_name not in self.levels:
            return {}
        
        level_data = {}
        for lump in self.levels[level_name]["lumps"]:
            if lump["name"] in ["VERTEXES", "LINEDEFS", "THINGS"]:
                level_data[lump["name"]] = self.get_lump_data(lump)
        
        return level_data
    
    @staticmethod
    def is_level_marker(name: str) -> bool:
        """Check if a lump name is a level marker."""
        # Doom 1 levels: E1M1 through E9M9
        if re.match(r'^E[1-9]M[1-9]$', name):
            return True
        
        # Doom 2 levels: MAP01 through MAP99
        if re.match(r'^MAP[0-9][0-9]$', name):
            return True
        
        return False
    
    def get_levels(self) -> List[str]:
        """Return a list of level names."""
        return list(self.levels.keys())
    
    def get_level_info(self, level_name: str) -> Dict:
        """Return information about a specific level."""
        return self.levels.get(level_name, {})
    
    @staticmethod
    def get_warp_params(level_name: str) -> Tuple[str, str]:
        """Convert a level name to warp parameters."""
        if level_name.startswith('E') and level_name[2] == 'M':
            # Doom 1 format: E1M1, E2M3, etc.
            episode = level_name[1]
            map_num = level_name[3]
            return episode, map_num
        elif level_name.startswith('MAP'):
            # Doom 2 format: MAP01, MAP02, etc.
            map_num = level_name[3:]
            return map_num, ""
        return "", ""
    
    @staticmethod
    def parse_vertices(data):
        """Parse VERTEXES lump. Each vertex is 4 bytes: two shorts (x, y)."""
        vertices = []
        for i in range(0, len(data), 4):
            x, y = struct.unpack_from("<hh", data, offset=i)
            vertices.append((x, y))
        return vertices
    
    @staticmethod
    def parse_linedefs(data):
        """Parse LINEDEFS lump. We only need start and end vertex indices."""
        linedefs = []
        for i in range(0, len(data), 14):
            values = struct.unpack_from("<7H", data, offset=i)
            start_vertex = values[0]
            end_vertex = values[1]
            linedefs.append((start_vertex, end_vertex))
        return linedefs
    
    @staticmethod
    def parse_things(data):
        """Parse THINGS lump. Each thing is 10 bytes."""
        things = []
        for i in range(0, len(data), 10):
            x, y, angle, typ, flags = struct.unpack_from("<5h", data, offset=i)
            things.append((x, y, angle, typ, flags))
        return things

class LevelPreviewDialog(Toplevel):
    def __init__(self, parent, wad_parser, level_name):
        super().__init__(parent)
        self.wad_parser = wad_parser
        self.level_name = level_name
        self.title(f"Level Preview: {level_name}")
        self.geometry("800x600")
        
        # Level data
        self.vertices = []
        self.linedefs = []
        self.things = []
        self.bounds = (0, 0, 0, 0)  # min_x, min_y, max_x, max_y
        
        # Configure window
        self.transient(parent)
        self.grab_set()
        self.protocol("WM_DELETE_WINDOW", self.on_close)
        
        # Create UI elements
        self.create_widgets()
        
        # Load and draw the level
        self.load_level()
        
        # Center the window
        self.update_idletasks()
        width = self.winfo_width()
        height = self.winfo_height()
        x = (self.winfo_screenwidth() // 2) - (width // 2)
        y = (self.winfo_screenheight() // 2) - (height // 2)
        self.geometry(f'{width}x{height}+{x}+{y}')
        
        # Auto-fit level to view when first opened
        self.after(100, self.fit_to_view)
    
    def create_widgets(self):
        """Create the UI elements for the preview dialog."""
        # Top frame for information
        top_frame = ttk.Frame(self, padding=5)
        top_frame.pack(fill=tk.X)
        
        # Level name and stats
        self.info_label = ttk.Label(top_frame, text="Loading level...")
        self.info_label.pack(side=tk.LEFT)
        
        # Canvas for map display with scrollbars
        canvas_frame = ttk.Frame(self)
        canvas_frame.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        self.canvas = tk.Canvas(canvas_frame, bg="black")
        self.canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        
        # Vertical scrollbar
        vscrollbar = ttk.Scrollbar(canvas_frame, orient=tk.VERTICAL, command=self.canvas.yview)
        vscrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        self.canvas.config(yscrollcommand=vscrollbar.set)
        
        # Horizontal scrollbar
        hscrollbar = ttk.Scrollbar(self, orient=tk.HORIZONTAL, command=self.canvas.xview)
        hscrollbar.pack(fill=tk.X)
        self.canvas.config(xscrollcommand=hscrollbar.set)
        
        # Legend and controls at the bottom
        bottom_frame = ttk.Frame(self, padding=5)
        bottom_frame.pack(fill=tk.X)
        
        # Legend
        legend_frame = ttk.Frame(bottom_frame)
        legend_frame.pack(side=tk.LEFT)
        
        ttk.Label(legend_frame, text="Legend:").pack(side=tk.LEFT)
        ttk.Label(legend_frame, text="■ Walls", foreground="white").pack(side=tk.LEFT, padx=5)
        ttk.Label(legend_frame, text="● Things", foreground="red").pack(side=tk.LEFT, padx=5)
        ttk.Label(legend_frame, text="× Player Start", foreground="green").pack(side=tk.LEFT, padx=5)
        
        # Controls
        control_frame = ttk.Frame(bottom_frame)
        control_frame.pack(side=tk.RIGHT)
        
        # Fit to view button
        self.fit_button = ttk.Button(control_frame, text="Fit to View", command=self.fit_to_view)
        self.fit_button.pack(side=tk.LEFT, padx=5)
        
        # Zoom controls
        zoom_frame = ttk.Frame(control_frame)
        zoom_frame.pack(side=tk.LEFT)
        
        ttk.Label(zoom_frame, text="Zoom:").pack(side=tk.LEFT, padx=5)
        self.zoom_scale = ttk.Scale(zoom_frame, from_=0.1, to=5.0, value=1.0, 
                                   orient=tk.HORIZONTAL, length=150,
                                   command=self.on_zoom_change)
        self.zoom_scale.pack(side=tk.LEFT)
        
        # Close button
        close_button = ttk.Button(control_frame, text="Close", command=self.on_close)
        close_button.pack(side=tk.LEFT, padx=10)
    
    def load_level(self):
        """Load the level data and draw it on the canvas."""
        # Get level lumps
        level_lumps = self.wad_parser.get_level_lumps(self.level_name)
        
        if not level_lumps:
            self.info_label.config(text=f"Error: Could not find level data for {self.level_name}")
            return
        
        # Check if required lumps are present
        if "VERTEXES" not in level_lumps or "LINEDEFS" not in level_lumps:
            self.info_label.config(text=f"Error: Missing required level data for {self.level_name}")
            return
        
        # Parse level data
        self.vertices = self.wad_parser.parse_vertices(level_lumps["VERTEXES"])
        self.linedefs = self.wad_parser.parse_linedefs(level_lumps["LINEDEFS"])
        self.things = []
        if "THINGS" in level_lumps:
            self.things = self.wad_parser.parse_things(level_lumps["THINGS"])
        
        # Update information label
        self.info_label.config(text=f"Level: {self.level_name} | Vertices: {len(self.vertices)} | "
                               f"Lines: {len(self.linedefs)} | Things: {len(self.things)}")
        
        # Calculate bounds
        self.calculate_bounds()
        
        # Draw the map
        self.draw_map()
    
    def calculate_bounds(self):
        """Calculate the bounding box of the level."""
        if not self.vertices and not self.things:
            self.bounds = (0, 0, 100, 100)  # Default if no data
            return
            
        # Get all X and Y coordinates
        all_xs = [v[0] for v in self.vertices]
        all_ys = [v[1] for v in self.vertices]
        
        if self.things:
            all_xs.extend([t[0] for t in self.things])
            all_ys.extend([t[1] for t in self.things])
        
        # Calculate bounds with margin
        margin = 100  # Add some padding around the level
        min_x = min(all_xs) - margin if all_xs else -margin
        min_y = min(all_ys) - margin if all_ys else -margin
        max_x = max(all_xs) + margin if all_xs else margin
        max_y = max(all_ys) + margin if all_ys else margin
        
        self.bounds = (min_x, min_y, max_x, max_y)
    
    def draw_map(self):
        """Draw the level map on the canvas."""
        self.canvas.delete("all")
        
        if not self.vertices and not self.things:
            return
        
        # Get canvas dimensions
        canvas_width = self.canvas.winfo_width()
        canvas_height = self.canvas.winfo_height()
        
        # Use default size if canvas hasn't been drawn yet
        if canvas_width <= 1:
            canvas_width = 800
        if canvas_height <= 1:
            canvas_height = 600
        
        # Get level bounds
        min_x, min_y, max_x, max_y = self.bounds
        level_width = max_x - min_x
        level_height = max_y - min_y
        
        # Get zoom factor
        zoom = self.zoom_scale.get()
        
        # Configure canvas scroll region (based on zoomed level size)
        zoomed_width = level_width * zoom
        zoomed_height = level_height * zoom
        self.canvas.config(scrollregion=(0, 0, zoomed_width, zoomed_height))
        
        def to_canvas(x, y):
            """Convert Doom coordinates to canvas coordinates."""
            # Normalize to origin (0,0) and apply zoom
            cx = (x - min_x) * zoom
            cy = (y - min_y) * zoom
            return cx, cy
        
        # Draw linedefs (walls)
        for start_idx, end_idx in self.linedefs:
            if start_idx < len(self.vertices) and end_idx < len(self.vertices):
                x1, y1 = self.vertices[start_idx]
                x2, y2 = self.vertices[end_idx]
                cx1, cy1 = to_canvas(x1, y1)
                cx2, cy2 = to_canvas(x2, y2)
                self.canvas.create_line(cx1, cy1, cx2, cy2, fill="white", width=1.5)
        
        # Draw things (items, monsters, etc.)
        for x, y, angle, typ, flags in self.things:
            cx, cy = to_canvas(x, y)
            
            # Player start (types 1-4) are green crosses
            if 1 <= typ <= 4:
                r = 6 * zoom
                self.canvas.create_line(cx-r, cy-r, cx+r, cy+r, fill="green", width=2)
                self.canvas.create_line(cx-r, cy+r, cx+r, cy-r, fill="green", width=2)
                
                # Draw direction indicator
                ang_rad = angle * (3.14159 / 180)
                dir_len = 10 * zoom
                dx = dir_len * -1 * (angle - 90) / 90  # Adjust for Doom's angle system
                dy = dir_len * (angle - 180) / 180
                self.canvas.create_line(cx, cy, cx + dx, cy + dy, fill="green", width=1.5)
            else:
                # Other things are red dots
                r = 3 * zoom
                self.canvas.create_oval(cx-r, cy-r, cx+r, cy+r, fill="red", outline="")
    
    def fit_to_view(self):
        """Automatically adjust zoom to fit the entire level in the canvas."""
        if not self.vertices and not self.things:
            return
            
        # Get canvas dimensions
        canvas_width = self.canvas.winfo_width()
        canvas_height = self.canvas.winfo_height()
        
        # Use default size if canvas hasn't been drawn yet
        if canvas_width <= 1:
            canvas_width = 800
        if canvas_height <= 1:
            canvas_height = 600
        
        # Get level bounds and size
        min_x, min_y, max_x, max_y = self.bounds
        level_width = max_x - min_x
        level_height = max_y - min_y
        
        # Calculate zoom factor to fit the level in the canvas
        # Use a margin of 20 pixels on each side
        margin = 40
        zoom_x = (canvas_width - margin) / level_width if level_width > 0 else 1
        zoom_y = (canvas_height - margin) / level_height if level_height > 0 else 1
        
        # Use the smaller zoom factor to ensure the entire level fits
        new_zoom = min(zoom_x, zoom_y)
        
        # Update the zoom scale
        self.zoom_scale.set(new_zoom)
        
        # Redraw with the new zoom
        self.draw_map()
        
        # Center the view
        self.canvas.xview_moveto(0)
        self.canvas.yview_moveto(0)
    
    def on_zoom_change(self, *args):
        """Redraw the map when zoom changes."""
        self.draw_map()
    
    def on_close(self):
        """Close the preview dialog."""
        self.grab_release()
        self.destroy()

class EngineConfigDialog(tk.Toplevel):
    def __init__(self, parent, engine_id=None, name="", path="", params=""):
        super().__init__(parent)
        self.parent = parent
        self.engine_id = engine_id
        self.result = None
        
        # Set window properties
        self.title("Engine Configuration")
        self.transient(parent)
        self.grab_set()
        self.geometry("500x250")
        
        # Create form fields
        frame = ttk.Frame(self, padding="10")
        frame.pack(fill=tk.BOTH, expand=True)
        
        # Engine name
        ttk.Label(frame, text="Engine Name:").grid(row=0, column=0, sticky=tk.W, padx=5, pady=5)
        self.name_var = tk.StringVar(value=name)
        ttk.Entry(frame, textvariable=self.name_var, width=40).grid(row=0, column=1, sticky=tk.W + tk.E, padx=5, pady=5)
        
        # Engine path
        ttk.Label(frame, text="Engine Path:").grid(row=1, column=0, sticky=tk.W, padx=5, pady=5)
        self.path_var = tk.StringVar(value=path)
        path_frame = ttk.Frame(frame)
        path_frame.grid(row=1, column=1, sticky=tk.W + tk.E, padx=5, pady=5)
        ttk.Entry(path_frame, textvariable=self.path_var, width=35).pack(side=tk.LEFT, fill=tk.X, expand=True)
        ttk.Button(path_frame, text="...", width=3, command=self.browse_engine).pack(side=tk.RIGHT, padx=2)
        
        # Default parameters
        ttk.Label(frame, text="Default Parameters:").grid(row=2, column=0, sticky=tk.W, padx=5, pady=5)
        self.params_var = tk.StringVar(value=params)
        ttk.Entry(frame, textvariable=self.params_var, width=40).grid(row=2, column=1, sticky=tk.W + tk.E, padx=5, pady=5)
        
        # Buttons
        button_frame = ttk.Frame(frame)
        button_frame.grid(row=3, column=0, columnspan=2, pady=10)
        ttk.Button(button_frame, text="Save", command=self.save).pack(side=tk.LEFT, padx=5)
        ttk.Button(button_frame, text="Cancel", command=self.cancel).pack(side=tk.LEFT, padx=5)
        
        # Configure resizing
        frame.columnconfigure(1, weight=1)
        
        # Center the window
        self.update_idletasks()
        width = self.winfo_width()
        height = self.winfo_height()
        x = (self.winfo_screenwidth() // 2) - (width // 2)
        y = (self.winfo_screenheight() // 2) - (height // 2)
        self.geometry(f'{width}x{height}+{x}+{y}')
        
        self.wait_window()
    
    def browse_engine(self):
        """Open a file dialog to select the Doom engine executable."""
        file_path = filedialog.askopenfilename(
            title="Select Doom Engine Executable",
            initialdir=os.path.dirname(self.path_var.get()) or os.path.expanduser("~"),
            filetypes=[("Executable files", "*.exe"), ("All files", "*.*")]
        )
        if file_path:
            self.path_var.set(file_path)
    
    def save(self):
        """Save the engine configuration."""
        name = self.name_var.get().strip()
        path = self.path_var.get().strip()
        
        if not name:
            messagebox.showerror("Error", "Engine name cannot be empty.")
            return
        
        if not path:
            messagebox.showerror("Error", "Engine path cannot be empty.")
            return
        
        if not os.path.isfile(path):
            if not messagebox.askyesno("Warning", "The specified engine executable does not exist. Save anyway?"):
                return
        
        self.result = {
            "id": self.engine_id or f"engine_{id(self)}",
            "name": name,
            "path": path,
            "params": self.params_var.get().strip()
        }
        self.destroy()
    
    def cancel(self):
        """Cancel the engine configuration."""
        self.result = None
        self.destroy()

class DirectoryConfigDialog(tk.Toplevel):
    def __init__(self, parent, engines, dir_id=None, name="", path="", engine_id="", params=""):
        super().__init__(parent)
        self.parent = parent
        self.engines = engines
        self.dir_id = dir_id
        self.result = None
        
        # Set window properties
        self.title("PWAD Directory Configuration")
        self.transient(parent)
        self.grab_set()
        self.geometry("500x280")
        
        # Create form fields
        frame = ttk.Frame(self, padding="10")
        frame.pack(fill=tk.BOTH, expand=True)
        
        # Directory name
        ttk.Label(frame, text="Directory Name:").grid(row=0, column=0, sticky=tk.W, padx=5, pady=5)
        self.name_var = tk.StringVar(value=name)
        ttk.Entry(frame, textvariable=self.name_var, width=40).grid(row=0, column=1, sticky=tk.W + tk.E, padx=5, pady=5)
        
        # Directory path
        ttk.Label(frame, text="Directory Path:").grid(row=1, column=0, sticky=tk.W, padx=5, pady=5)
        self.path_var = tk.StringVar(value=path)
        path_frame = ttk.Frame(frame)
        path_frame.grid(row=1, column=1, sticky=tk.W + tk.E, padx=5, pady=5)
        ttk.Entry(path_frame, textvariable=self.path_var, width=35).pack(side=tk.LEFT, fill=tk.X, expand=True)
        ttk.Button(path_frame, text="...", width=3, command=self.browse_directory).pack(side=tk.RIGHT, padx=2)
        
        # Associated engine
        ttk.Label(frame, text="Doom Engine:").grid(row=2, column=0, sticky=tk.W, padx=5, pady=5)
        
        # Create a mapping of engine IDs to names
        self.engine_id_to_name = {}
        self.engine_name_to_id = {}
        engine_names = []
        
        for engine in engines:
            engine_id = engine["id"]
            engine_name = engine["name"]
            self.engine_id_to_name[engine_id] = engine_name
            self.engine_name_to_id[engine_name] = engine_id
            engine_names.append(engine_name)
        
        # Create StringVar to hold the selected engine name (not ID)
        self.engine_var = tk.StringVar()
        
        # Set the initial engine name if engine_id is provided
        if engine_id and engine_id in self.engine_id_to_name:
            self.engine_var.set(self.engine_id_to_name[engine_id])
        elif engine_names:
            self.engine_var.set(engine_names[0])
        
        # Create the OptionMenu with engine names
        self.engine_dropdown = ttk.OptionMenu(frame, self.engine_var, self.engine_var.get(), *engine_names)
        self.engine_dropdown.grid(row=2, column=1, sticky=tk.W + tk.E, padx=5, pady=5)
        
        # Additional parameters
        ttk.Label(frame, text="Additional Parameters:").grid(row=3, column=0, sticky=tk.W, padx=5, pady=5)
        self.params_var = tk.StringVar(value=params)
        ttk.Entry(frame, textvariable=self.params_var, width=40).grid(row=3, column=1, sticky=tk.W + tk.E, padx=5, pady=5)
        
        # Buttons
        button_frame = ttk.Frame(frame)
        button_frame.grid(row=4, column=0, columnspan=2, pady=10)
        ttk.Button(button_frame, text="Save", command=self.save).pack(side=tk.LEFT, padx=5)
        ttk.Button(button_frame, text="Cancel", command=self.cancel).pack(side=tk.LEFT, padx=5)
        
        # Configure resizing
        frame.columnconfigure(1, weight=1)
        
        # Center the window
        self.update_idletasks()
        width = self.winfo_width()
        height = self.winfo_height()
        x = (self.winfo_screenwidth() // 2) - (width // 2)
        y = (self.winfo_screenheight() // 2) - (height // 2)
        self.geometry(f'{width}x{height}+{x}+{y}')
        
        self.wait_window()
    
    def browse_directory(self):
        """Open a directory dialog to select the PWAD parent directory."""
        directory = filedialog.askdirectory(
            title="Select PWAD Parent Directory",
            initialdir=self.path_var.get() or os.path.expanduser("~")
        )
        if directory:
            self.path_var.set(directory)
    
    def save(self):
        """Save the directory configuration."""
        name = self.name_var.get().strip()
        path = self.path_var.get().strip()
        engine_name = self.engine_var.get()
        
        if not name:
            messagebox.showerror("Error", "Directory name cannot be empty.")
            return
        
        if not path:
            messagebox.showerror("Error", "Directory path cannot be empty.")
            return
        
        if not os.path.isdir(path):
            if not messagebox.askyesno("Warning", "The specified directory does not exist. Save anyway?"):
                return
        
        if not engine_name:
            messagebox.showerror("Error", "Please select an engine.")
            return
        
        # Convert engine name to ID
        engine_id = self.engine_name_to_id.get(engine_name)
        if not engine_id:
            messagebox.showerror("Error", "Invalid engine selection.")
            return
        
        self.result = {
            "id": self.dir_id or f"dir_{id(self)}",
            "name": name,
            "path": path,
            "engine_id": engine_id,
            "params": self.params_var.get().strip()
        }
        self.destroy()
    
    def cancel(self):
        """Cancel the directory configuration."""
        self.result = None
        self.destroy()

class EngineManagerDialog(tk.Toplevel):
    def __init__(self, parent, engines):
        super().__init__(parent)
        self.parent = parent
        self.engines = engines.copy()
        self.result = None
        
        # Set window properties
        self.title("Doom Engine Manager")
        self.transient(parent)
        self.grab_set()
        self.geometry("600x400")
        
        # Create main frame
        main_frame = ttk.Frame(self, padding="10")
        main_frame.pack(fill=tk.BOTH, expand=True)
        
        # Engine listbox with scrollbar
        list_frame = ttk.LabelFrame(main_frame, text="Available Engines")
        list_frame.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        list_box_frame = ttk.Frame(list_frame)
        list_box_frame.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        # Columns for the treeview
        columns = ("name", "path")
        self.engine_tree = ttk.Treeview(list_box_frame, columns=columns, show="headings")
        self.engine_tree.heading("name", text="Engine Name")
        self.engine_tree.heading("path", text="Path")
        self.engine_tree.column("name", width=150)
        self.engine_tree.column("path", width=350)
        self.engine_tree.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        
        scrollbar = ttk.Scrollbar(list_box_frame, orient=tk.VERTICAL, command=self.engine_tree.yview)
        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        self.engine_tree.config(yscrollcommand=scrollbar.set)
        
        # Buttons for add/edit/remove
        button_frame = ttk.Frame(main_frame)
        button_frame.pack(fill=tk.X, padx=5, pady=5)
        
        ttk.Button(button_frame, text="Add Engine", command=self.add_engine).pack(side=tk.LEFT, padx=5)
        ttk.Button(button_frame, text="Edit Engine", command=self.edit_engine).pack(side=tk.LEFT, padx=5)
        ttk.Button(button_frame, text="Remove Engine", command=self.remove_engine).pack(side=tk.LEFT, padx=5)
        
        # Close button
        close_frame = ttk.Frame(main_frame)
        close_frame.pack(fill=tk.X, padx=5, pady=10)
        
        ttk.Button(close_frame, text="Save Changes", command=self.save_changes).pack(side=tk.RIGHT, padx=5)
        ttk.Button(close_frame, text="Cancel", command=self.cancel).pack(side=tk.RIGHT, padx=5)
        
        # Populate the listbox
        self.populate_engine_list()
        
        # Center the window
        self.update_idletasks()
        width = self.winfo_width()
        height = self.winfo_height()
        x = (self.winfo_screenwidth() // 2) - (width // 2)
        y = (self.winfo_screenheight() // 2) - (height // 2)
        self.geometry(f'{width}x{height}+{x}+{y}')
        
        self.wait_window()
    
    def populate_engine_list(self):
        """Populate the engine treeview with the current engines."""
        self.engine_tree.delete(*self.engine_tree.get_children())
        
        for engine in self.engines:
            self.engine_tree.insert("", "end", values=(engine["name"], engine["path"]), tags=(engine["id"],))
    
    def add_engine(self):
        """Add a new engine."""
        dialog = EngineConfigDialog(self)
        if dialog.result:
            self.engines.append(dialog.result)
            self.populate_engine_list()
    
    def edit_engine(self):
        """Edit the selected engine."""
        selection = self.engine_tree.selection()
        if not selection:
            messagebox.showinfo("Info", "Please select an engine to edit.")
            return
        
        # Get the engine ID from the tags
        engine_id = self.engine_tree.item(selection[0], "tags")[0]
        
        # Find the engine in our list
        engine = next((e for e in self.engines if e["id"] == engine_id), None)
        if not engine:
            return
        
        dialog = EngineConfigDialog(
            self,
            engine_id=engine["id"],
            name=engine["name"],
            path=engine["path"],
            params=engine.get("params", "")
        )
        
        if dialog.result:
            # Replace the engine in our list
            for i, e in enumerate(self.engines):
                if e["id"] == engine_id:
                    self.engines[i] = dialog.result
                    break
            
            self.populate_engine_list()
    
    def remove_engine(self):
        """Remove the selected engine."""
        selection = self.engine_tree.selection()
        if not selection:
            messagebox.showinfo("Info", "Please select an engine to remove.")
            return
        
        # Get the engine ID from the tags
        engine_id = self.engine_tree.item(selection[0], "tags")[0]
        
        # Confirm deletion
        if not messagebox.askyesno("Confirm", "Are you sure you want to remove this engine?"):
            return
        
        # Remove the engine from our list
        self.engines = [e for e in self.engines if e["id"] != engine_id]
        self.populate_engine_list()
    
    def save_changes(self):
        """Save changes and close the dialog."""
        self.result = self.engines
        self.destroy()
    
    def cancel(self):
        """Cancel changes and close the dialog."""
        self.result = None
        self.destroy()

class DirectoryManagerDialog(tk.Toplevel):
    def __init__(self, parent, directories, engines):
        super().__init__(parent)
        self.parent = parent
        self.directories = directories.copy()
        self.engines = engines
        self.result = None
        
        # Set window properties
        self.title("PWAD Directory Manager")
        self.transient(parent)
        self.grab_set()
        self.geometry("600x400")
        
        # Create main frame
        main_frame = ttk.Frame(self, padding="10")
        main_frame.pack(fill=tk.BOTH, expand=True)
        
        # Directory listbox with scrollbar
        list_frame = ttk.LabelFrame(main_frame, text="PWAD Directories")
        list_frame.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        list_box_frame = ttk.Frame(list_frame)
        list_box_frame.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        # Columns for the treeview
        columns = ("name", "path", "engine")
        self.dir_tree = ttk.Treeview(list_box_frame, columns=columns, show="headings")
        self.dir_tree.heading("name", text="Directory Name")
        self.dir_tree.heading("path", text="Path")
        self.dir_tree.heading("engine", text="Engine")
        self.dir_tree.column("name", width=150)
        self.dir_tree.column("path", width=250)
        self.dir_tree.column("engine", width=100)
        self.dir_tree.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        
        scrollbar = ttk.Scrollbar(list_box_frame, orient=tk.VERTICAL, command=self.dir_tree.yview)
        scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        self.dir_tree.config(yscrollcommand=scrollbar.set)
        
        # Buttons for add/edit/remove
        button_frame = ttk.Frame(main_frame)
        button_frame.pack(fill=tk.X, padx=5, pady=5)
        
        ttk.Button(button_frame, text="Add Directory", command=self.add_directory).pack(side=tk.LEFT, padx=5)
        ttk.Button(button_frame, text="Edit Directory", command=self.edit_directory).pack(side=tk.LEFT, padx=5)
        ttk.Button(button_frame, text="Remove Directory", command=self.remove_directory).pack(side=tk.LEFT, padx=5)
        
        # Close button
        close_frame = ttk.Frame(main_frame)
        close_frame.pack(fill=tk.X, padx=5, pady=10)
        
        ttk.Button(close_frame, text="Save Changes", command=self.save_changes).pack(side=tk.RIGHT, padx=5)
        ttk.Button(close_frame, text="Cancel", command=self.cancel).pack(side=tk.RIGHT, padx=5)
        
        # Populate the listbox
        self.populate_directory_list()
        
        # Center the window
        self.update_idletasks()
        width = self.winfo_width()
        height = self.winfo_height()
        x = (self.winfo_screenwidth() // 2) - (width // 2)
        y = (self.winfo_screenheight() // 2) - (height // 2)
        self.geometry(f'{width}x{height}+{x}+{y}')
        
        self.wait_window()
    
    def get_engine_name(self, engine_id):
        """Get the name of an engine by its ID."""
        for engine in self.engines:
            if engine["id"] == engine_id:
                return engine["name"]
        return "Unknown Engine"
    
    def populate_directory_list(self):
        """Populate the directory treeview with the current directories."""
        self.dir_tree.delete(*self.dir_tree.get_children())
        
        for directory in self.directories:
            engine_name = self.get_engine_name(directory["engine_id"])
            self.dir_tree.insert("", "end", values=(directory["name"], directory["path"], engine_name), tags=(directory["id"],))
    
    def add_directory(self):
        """Add a new directory."""
        if not self.engines:
            messagebox.showinfo("Info", "Please add at least one engine first.")
            return
            
        dialog = DirectoryConfigDialog(self, self.engines)
        if dialog.result:
            self.directories.append(dialog.result)
            self.populate_directory_list()
    
    def edit_directory(self):
        """Edit the selected directory."""
        selection = self.dir_tree.selection()
        if not selection:
            messagebox.showinfo("Info", "Please select a directory to edit.")
            return
        
        # Get the directory ID from the tags
        dir_id = self.dir_tree.item(selection[0], "tags")[0]
        
        # Find the directory in our list
        directory = next((d for d in self.directories if d["id"] == dir_id), None)
        if not directory:
            return
        
        dialog = DirectoryConfigDialog(
            self,
            self.engines,
            dir_id=directory["id"],
            name=directory["name"],
            path=directory["path"],
            engine_id=directory["engine_id"],
            params=directory.get("params", "")
        )
        
        if dialog.result:
            # Replace the directory in our list
            for i, d in enumerate(self.directories):
                if d["id"] == dir_id:
                    self.directories[i] = dialog.result
                    break
            
            self.populate_directory_list()
    
    def remove_directory(self):
        """Remove the selected directory."""
        selection = self.dir_tree.selection()
        if not selection:
            messagebox.showinfo("Info", "Please select a directory to remove.")
            return
        
        # Get the directory ID from the tags
        dir_id = self.dir_tree.item(selection[0], "tags")[0]
        
        # Confirm deletion
        if not messagebox.askyesno("Confirm", "Are you sure you want to remove this directory?"):
            return
        
        # Remove the directory from our list
        self.directories = [d for d in self.directories if d["id"] != dir_id]
        self.populate_directory_list()
    
    def save_changes(self):
        """Save changes and close the dialog."""
        self.result = self.directories
        self.destroy()
    
    def cancel(self):
        """Cancel changes and close the dialog."""
        self.result = None
        self.destroy()

class PWADManager:
    def __init__(self, root):
        self.root = root
        self.root.title("Doom PWAD Manager")
        self.root.geometry("900x650")
        
        # Initialize variables
        self.engines = []
        self.directories = []
        self.current_directory = None
        self.selected_difficulty = tk.StringVar(value="Hurt me plenty (3)")
        self.selected_level = tk.StringVar()
        self.current_pwad = None
        self.current_wad_parser = None
        
        # Initialize config
        self.config_file = "config.json"
        self.load_config()
        
        # Setup UI
        self.setup_menu()
        self.setup_ui()
        
        # Populate directory selector
        self.update_directory_selector()
    
    def setup_menu(self):
        """Set up the application menu."""
        menubar = tk.Menu(self.root)
        
        # File menu
        file_menu = tk.Menu(menubar, tearoff=0)
        file_menu.add_command(label="Manage Engines", command=self.manage_engines)
        file_menu.add_command(label="Manage Directories", command=self.manage_directories)
        file_menu.add_separator()
        file_menu.add_command(label="Exit", command=self.root.quit)
        menubar.add_cascade(label="File", menu=file_menu)
        
        self.root.config(menu=menubar)
    
    def setup_ui(self):
        """Set up the main UI components."""
        # Create main frame with padding
        main_frame = ttk.Frame(self.root, padding="10")
        main_frame.pack(fill=tk.BOTH, expand=True)
        
        # Directory selector
        selector_frame = ttk.Frame(main_frame)
        selector_frame.pack(fill=tk.X, padx=5, pady=5)
        
        ttk.Label(selector_frame, text="PWAD Directory:").pack(side=tk.LEFT, padx=5)
        
        self.directory_var = tk.StringVar()
        self.directory_dropdown = ttk.Combobox(selector_frame, textvariable=self.directory_var, state="readonly", width=40)
        self.directory_dropdown.pack(side=tk.LEFT, padx=5, fill=tk.X, expand=True)
        self.directory_dropdown.bind("<<ComboboxSelected>>", self.on_directory_select)
        
        # Left pane: PWAD list
        left_frame = ttk.LabelFrame(main_frame, text="PWAD Directories")
        left_frame.pack(side=tk.LEFT, fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        # PWAD listbox with scrollbar
        pwad_frame = ttk.Frame(left_frame)
        pwad_frame.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        self.pwad_listbox = tk.Listbox(pwad_frame)
        self.pwad_listbox.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        
        pwad_scrollbar = ttk.Scrollbar(pwad_frame, orient=tk.VERTICAL, command=self.pwad_listbox.yview)
        pwad_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        self.pwad_listbox.config(yscrollcommand=pwad_scrollbar.set)
        
        # Bind selection event
        self.pwad_listbox.bind('<<ListboxSelect>>', self.on_pwad_select)
        
        # Right pane: PWAD details, level selector, and launch options
        right_frame = ttk.Frame(main_frame)
        right_frame.pack(side=tk.RIGHT, fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        # Engine info
        engine_frame = ttk.Frame(right_frame)
        engine_frame.pack(fill=tk.X, padx=5, pady=5)
        
        self.engine_label = ttk.Label(engine_frame, text="Engine: None")
        self.engine_label.pack(side=tk.LEFT, padx=5)
        
        # PWAD details
        details_frame = ttk.LabelFrame(right_frame, text="PWAD Details")
        details_frame.pack(fill=tk.BOTH, expand=True, padx=5, pady=5)
        
        self.details_text = tk.Text(details_frame, wrap=tk.WORD, height=15)
        self.details_text.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        
        details_scrollbar = ttk.Scrollbar(details_frame, orient=tk.VERTICAL, command=self.details_text.yview)
        details_scrollbar.pack(side=tk.RIGHT, fill=tk.Y)
        self.details_text.config(yscrollcommand=details_scrollbar.set)
        
        # Launch options
        options_frame = ttk.LabelFrame(right_frame, text="Launch Options")
        options_frame.pack(fill=tk.X, padx=5, pady=5)
        
        # Level selector
        level_frame = ttk.Frame(options_frame)
        level_frame.pack(fill=tk.X, padx=5, pady=2)
        
        self.level_label = ttk.Label(level_frame, text="Level:")
        self.level_label.pack(side=tk.LEFT, padx=5)
        
        self.level_menu = ttk.OptionMenu(level_frame, self.selected_level, "")
        self.level_menu.pack(side=tk.LEFT, fill=tk.X, expand=True, padx=5)
        
        # "Preview Level" button
        self.preview_button = ttk.Button(level_frame, text="Preview", command=self.preview_level, state="disabled")
        self.preview_button.pack(side=tk.RIGHT, padx=5)
        
        # Difficulty selector
        difficulty_frame = ttk.Frame(options_frame)
        difficulty_frame.pack(fill=tk.X, padx=5, pady=2)
        
        ttk.Label(difficulty_frame, text="Difficulty:").pack(side=tk.LEFT, padx=5)
        difficulties = [
            "I'm too young to die (1)",
            "Hey, not too rough (2)",
            "Hurt me plenty (3)",
            "Ultra-Violence (4)",
            "Nightmare! (5)"
        ]
        self.difficulty_menu = ttk.OptionMenu(
            difficulty_frame, 
            self.selected_difficulty, 
            self.selected_difficulty.get(),
            *difficulties,
            command=self.save_difficulty
        )
        self.difficulty_menu.pack(side=tk.LEFT, fill=tk.X, expand=True, padx=5)
        
        # Launch buttons
        launch_frame = ttk.Frame(right_frame)
        launch_frame.pack(fill=tk.X, padx=5, pady=10)
        
        # Standard launch button (with selected level)
        self.launch_button = ttk.Button(launch_frame, text="Launch PWAD", command=self.launch_pwad)
        self.launch_button.pack(padx=5, pady=5, fill=tk.X)
        
        # Launch without warp button
        self.launch_no_warp_button = ttk.Button(
            launch_frame, 
            text="Launch PWAD without warp", 
            command=self.launch_pwad_no_warp
        )
        self.launch_no_warp_button.pack(padx=5, pady=5, fill=tk.X)
        
        # Status bar
        status_frame = ttk.Frame(self.root)
        status_frame.pack(fill=tk.X, side=tk.BOTTOM)
        
        self.status_label = ttk.Label(status_frame, text="Ready", anchor=tk.W, relief=tk.SUNKEN)
        self.status_label.pack(fill=tk.X)
    
    def update_directory_selector(self):
        """Update the directory dropdown with the available directories."""
        self.directory_dropdown["values"] = [d["name"] for d in self.directories]
        
        if self.directories and not self.directory_var.get():
            self.directory_var.set(self.directories[0]["name"])
            self.on_directory_select(None)
    
    def on_directory_select(self, event):
        """Handle selection of a directory from the dropdown."""
        selected_name = self.directory_var.get()
        
        # Find the directory by name
        directory = next((d for d in self.directories if d["name"] == selected_name), None)
        if not directory:
            return
        
        self.current_directory = directory
        self.populate_pwad_list()
        
        # Update engine label
        engine = self.get_engine_by_id(directory["engine_id"])
        if engine:
            self.engine_label.config(text=f"Engine: {engine['name']}")
        else:
            self.engine_label.config(text="Engine: Unknown")
    
    def get_engine_by_id(self, engine_id):
        """Get an engine by its ID."""
        return next((e for e in self.engines if e["id"] == engine_id), None)
    
    def manage_engines(self):
        """Open the engine manager dialog."""
        dialog = EngineManagerDialog(self.root, self.engines)
        if dialog.result is not None:
            self.engines = dialog.result
            self.save_config()
            
            # Update the UI
            if self.current_directory:
                engine = self.get_engine_by_id(self.current_directory["engine_id"])
                if engine:
                    self.engine_label.config(text=f"Engine: {engine['name']}")
                else:
                    self.engine_label.config(text="Engine: Unknown")
    
    def manage_directories(self):
        """Open the directory manager dialog."""
        dialog = DirectoryManagerDialog(self.root, self.directories, self.engines)
        if dialog.result is not None:
            self.directories = dialog.result
            self.save_config()
            
            # Update the UI
            self.update_directory_selector()
    
    def populate_pwad_list(self):
        """Populate the PWAD listbox with directories from the selected PWAD parent directory."""
        self.pwad_listbox.delete(0, tk.END)
        
        if not self.current_directory or not os.path.isdir(self.current_directory["path"]):
            return
        
        try:
            # Get subdirectories and sort them alphabetically
            pwad_dirs = []
            for entry in os.scandir(self.current_directory["path"]):
                if entry.is_dir():
                    pwad_dirs.append(entry.name)
            
            pwad_dirs.sort(key=str.lower)
            
            # Add to listbox
            for dir_name in pwad_dirs:
                self.pwad_listbox.insert(tk.END, dir_name)
            
            self.status_label.config(text=f"Found {len(pwad_dirs)} PWAD directories")
        except Exception as e:
            messagebox.showerror("Error", f"Failed to read PWAD directory: {e}")
    
    def find_text_files(self, directory: str) -> List[str]:
        """Find text files in a directory, prioritizing .txt files."""
        txt_files = []
        diz_files = []
        
        for file in os.listdir(directory):
            lower_file = file.lower()
            if lower_file.endswith('.txt'):
                txt_files.append(file)
            elif lower_file.endswith('.diz'):
                diz_files.append(file)
        
        # Return txt files first, then diz files
        return txt_files + diz_files
    
    def on_pwad_select(self, event):
        """Handle selection of a PWAD from the listbox."""
        selection = self.pwad_listbox.curselection()
        if not selection:
            return
        
        pwad_name = self.pwad_listbox.get(selection[0])
        pwad_path = os.path.join(self.current_directory["path"], pwad_name)
        self.current_pwad = pwad_path
        
        self.details_text.delete(1.0, tk.END)
        self.selected_level.set("")
        self.preview_button.config(state="disabled")  # Disable preview button until a valid level is selected
        
        # Clear level dropdown
        self.level_menu['menu'].delete(0, 'end')
        
        # Check if directory exists
        if not os.path.isdir(pwad_path):
            self.details_text.insert(tk.END, f"Directory not found: {pwad_path}")
            return
        
        # Look for text files, prioritizing .txt over .diz
        text_content = ""
        text_files = self.find_text_files(pwad_path)
        
        if text_files:
            for file in text_files:
                try:
                    with open(os.path.join(pwad_path, file), 'r', errors='replace') as f:
                        text_content += f"=== {file} ===\n\n"
                        text_content += f.read() + "\n\n"
                except Exception as e:
                    text_content += f"Error reading {file}: {e}\n\n"
        
        # Look for WAD files
        wad_files = [f for f in os.listdir(pwad_path) if f.lower().endswith('.wad')]
        
        if wad_files:
            text_content += f"Found WAD files: {', '.join(wad_files)}\n\n"
            
            # Parse the first WAD file
            wad_path = os.path.join(pwad_path, wad_files[0])
            try:
                self.current_wad_parser = WADParser(wad_path)
                levels = self.current_wad_parser.get_levels()
                
                if levels:
                    text_content += f"Levels found: {', '.join(levels)}\n\n"
                    
                    # Update the level label with count
                    self.level_label.config(text=f"Level ({len(levels)}):")
                    
                    # Add levels to dropdown
                    for level in levels:
                        self.level_menu['menu'].add_command(
                            label=level,
                            command=lambda lev=level: self.select_level(lev)
                        )
                    
                    # Select the first level by default
                    self.selected_level.set(levels[0])
                    self.preview_button.config(state="normal")  # Enable preview button
                    
                    # Add level details
                    for level in levels:
                        level_info = self.current_wad_parser.get_level_info(level)
                        text_content += f"Level {level}:\n"
                        lumps = [lump["name"] for lump in level_info.get('lumps', [])]
                        text_content += f"  Lumps: {', '.join(lumps)}\n"
                        if 'thing_count' in level_info:
                            text_content += f"  Things: {level_info['thing_count']}\n"
                        text_content += "\n"
                else:
                    text_content += "No levels detected (WAD can still be launched).\n"
                    self.level_label.config(text="No levels detected (WAD can still be launched)")
            except Exception as e:
                text_content += f"Error parsing WAD file: {e}\n"
                self.level_label.config(text="Error parsing WAD file (WAD can still be launched)")
        else:
            text_content += "No WAD files found in directory.\n"
            self.level_label.config(text="No WAD files found")
        
        self.details_text.insert(tk.END, text_content)
    
    def select_level(self, level):
        """Set the selected level and enable preview button."""
        self.selected_level.set(level)
        self.preview_button.config(state="normal")
    
    def preview_level(self):
        """Open a preview window for the selected level."""
        if not self.current_wad_parser or not self.selected_level.get():
            messagebox.showinfo("Info", "No level selected to preview.")
            return
        
        # Create and show the level preview dialog
        LevelPreviewDialog(self.root, self.current_wad_parser, self.selected_level.get())
    
    def launch_pwad(self):
        """Launch the selected PWAD with the configured Doom engine."""
        if not self.current_directory:
            messagebox.showinfo("Info", "Please select a PWAD directory first.")
            return
        
        # Get the engine
        engine = self.get_engine_by_id(self.current_directory["engine_id"])
        if not engine:
            messagebox.showerror("Error", "The engine associated with this directory is not available.")
            return
        
        # Validate engine path
        if not os.path.isfile(engine["path"]):
            messagebox.showerror("Error", f"Engine executable not found: {engine['path']}")
            return
        
        # Get selected PWAD
        selection = self.pwad_listbox.curselection()
        if not selection:
            messagebox.showinfo("Info", "Please select a PWAD from the list.")
            return
        
        pwad_name = self.pwad_listbox.get(selection[0])
        pwad_path = os.path.join(self.current_directory["path"], pwad_name)
        
        # Find WAD file
        wad_files = [f for f in os.listdir(pwad_path) if f.lower().endswith('.wad')]
        if not wad_files:
            messagebox.showinfo("Info", "No WAD files found in the selected directory.")
            return
        
        wad_file = os.path.join(pwad_path, wad_files[0])
        
        # Start building the command with the engine path
        cmd = [engine["path"]]
        
        # Add engine default parameters
        if engine.get("params", ""):
            cmd.extend(engine["params"].split())
        
        # Add directory-specific parameters
        if self.current_directory.get("params", ""):
            cmd.extend(self.current_directory["params"].split())
        
        # Add the WAD file
        cmd.extend(["-file", wad_file])
        
        # Add warp parameter if a level is selected
        if self.selected_level.get():
            ep, m = WADParser.get_warp_params(self.selected_level.get())
            if ep and m:  # Doom 1 format
                cmd.extend(["-warp", ep, m])
            elif ep:  # Doom 2 format
                cmd.extend(["-warp", ep])
        
        # Add skill parameter
        difficulty_mapping = {
            "I'm too young to die (1)": "1",
            "Hey, not too rough (2)": "2",
            "Hurt me plenty (3)": "3",
            "Ultra-Violence (4)": "4",
            "Nightmare! (5)": "5"
        }
        
        if self.selected_difficulty.get() in difficulty_mapping:
            cmd.extend(["-skill", difficulty_mapping[self.selected_difficulty.get()]])
        
        # Launch the game
        try:
            subprocess.Popen(cmd)
            self.status_label.config(text=f"Launched: {' '.join(cmd)}")
        except Exception as e:
            messagebox.showerror("Error", f"Failed to launch game: {e}")
    
    def launch_pwad_no_warp(self):
        """Launch the selected PWAD without any warp parameters."""
        if not self.current_directory:
            messagebox.showinfo("Info", "Please select a PWAD directory first.")
            return
        
        # Get the engine
        engine = self.get_engine_by_id(self.current_directory["engine_id"])
        if not engine:
            messagebox.showerror("Error", "The engine associated with this directory is not available.")
            return
        
        # Validate engine path
        if not os.path.isfile(engine["path"]):
            messagebox.showerror("Error", f"Engine executable not found: {engine['path']}")
            return
        
        # Get selected PWAD
        selection = self.pwad_listbox.curselection()
        if not selection:
            messagebox.showinfo("Info", "Please select a PWAD from the list.")
            return
        
        pwad_name = self.pwad_listbox.get(selection[0])
        pwad_path = os.path.join(self.current_directory["path"], pwad_name)
        
        # Find WAD file
        wad_files = [f for f in os.listdir(pwad_path) if f.lower().endswith('.wad')]
        if not wad_files:
            messagebox.showinfo("Info", "No WAD files found in the selected directory.")
            return
        
        wad_file = os.path.join(pwad_path, wad_files[0])
        
        # Start building the command with the engine path
        cmd = [engine["path"]]
        
        # Add engine default parameters
        if engine.get("params", ""):
            cmd.extend(engine["params"].split())
        
        # Add directory-specific parameters
        if self.current_directory.get("params", ""):
            cmd.extend(self.current_directory["params"].split())
        
        # Add the WAD file
        cmd.extend(["-file", wad_file])
        
        # Add skill parameter only (no warp)
        difficulty_mapping = {
            "I'm too young to die (1)": "1",
            "Hey, not too rough (2)": "2",
            "Hurt me plenty (3)": "3",
            "Ultra-Violence (4)": "4",
            "Nightmare! (5)": "5"
        }
        
        if self.selected_difficulty.get() in difficulty_mapping:
            cmd.extend(["-skill", difficulty_mapping[self.selected_difficulty.get()]])
        
        # Launch the game
        try:
            subprocess.Popen(cmd)
            self.status_label.config(text=f"Launched without warp: {' '.join(cmd)}")
        except Exception as e:
            messagebox.showerror("Error", f"Failed to launch game: {e}")
    
    def save_difficulty(self, *args):
        """Save the selected difficulty to the config."""
        self.save_config()
    
    def load_config(self):
        """Load configuration from the config file."""
        try:
            if os.path.isfile(self.config_file):
                with open(self.config_file, 'r') as f:
                    config = json.load(f)
                
                self.engines = config.get('engines', [])
                self.directories = config.get('directories', [])
                self.selected_difficulty.set(config.get('difficulty', 'Hurt me plenty (3)'))
            else:
                # Create default engine and directory if no config exists
                self.create_default_config()
        except Exception as e:
            messagebox.showerror("Error", f"Failed to load configuration: {e}")
            self.create_default_config()
    
    def create_default_config(self):
        """Create a default configuration."""
        # Default engine (GZDoom)
        default_engine = {
            "id": "default_engine",
            "name": "GZDoom",
            "path": "",
            "params": "-iwad doom2.wad"
        }
        self.engines = [default_engine]
        
        # Default directory
        default_directory = {
            "id": "default_directory",
            "name": "Default PWADs",
            "path": "",
            "engine_id": default_engine["id"],
            "params": ""
        }
        self.directories = [default_directory]
        
        # Save the default config
        self.save_config()
    
    def save_config(self):
        """Save configuration to the config file."""
        try:
            config = {
                'engines': self.engines,
                'directories': self.directories,
                'difficulty': self.selected_difficulty.get()
            }
            
            with open(self.config_file, 'w') as f:
                json.dump(config, f, indent=4)
        except Exception as e:
            messagebox.showerror("Error", f"Failed to save configuration: {e}")

def main():
    root = tk.Tk()
    app = PWADManager(root)
    root.mainloop()

if __name__ == "__main__":
    main()